@@ -29,5 +29,5 @@ logging.basicConfig(
     else "notset",
     format="%(message)s",
     datefmt="[%X]",
-    handlers=[RichHandler(console=console)],
+    handlers=[RichHandler(console=console, rich_tracebacks=True)],
 )
