@@ -7,15 +7,17 @@ import shlex
 import pathlib
 import re
 import json
-import collections
+from collections import defaultdict, Counter
 import threading
 import time
+import pkg_resources
+import functools
+import itertools
 
 # internal modules
 from thunar_plugins import l10n
 from thunar_plugins.log import console, Notify
 from thunar_plugins.version import __version__
-from thunar_plugins.log import console, Notify
 
 # external modules
 from rich.pretty import Pretty
@@ -23,58 +25,101 @@ import gi
 
 gi.require_version("Gtk", "3.0")
 
-from gi.repository import GObject, Gtk, Thunarx
+from gi.repository import GObject, Gtk, GLib, Thunarx
 
 logger = logging.getLogger(__name__)
 
 
-METADATA_CACHE = collections.defaultdict(lambda: collections.defaultdict(set))
+METADATA_CACHE = defaultdict(lambda: defaultdict(set))
 
 
-class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
-    def __init__(self):
-        logger.debug(f"GitAnnexSubmenu is initialized")
-
-    @property
-    def counter(self):
-        try:
-            return self._counter
-        except AttributeError:
-            self._counter = collections.Counter()
-        return self._counter
+class GitAnnexHelper:
+    @classmethod
+    def flush_gtk(cls):
+        while Gtk.events_pending():
+            Gtk.main_iteration()
 
     @classmethod
-    def name(cls):
-        s = _("Git Annex Context-Menu")
-        if not shutil.which("git-annex"):
-            s = _("[Unavailable]") + " " + s
-        return s
+    def not_in_parallel(cls, decorated_fun):
+        if not hasattr(cls, "locks"):
+            cls.locks = dict()
+
+        cls.locks[decorated_fun] = threading.Lock()
+
+        @functools.wraps(decorated_fun)
+        def wrapper(*args, **kwargs):
+            if (lock := cls.locks[decorated_fun]).locked():
+                logger.debug(
+                    f"Not running {decorated_fun}(*{args!r},**{kwargs!r}) as it is already running!"
+                )
+            else:
+                with lock:
+                    decorated_fun(*args, **kwargs)
+
+        return wrapper
 
     @classmethod
-    def description(cls):
-        s = _(
-            "This plugin adds a context menu item "
-            "for managing Git Annex repositories."
-        )
-        if not shutil.which("git-annex"):
-            s += " " + _(
-                "Install Git Annex to use this plugin: "
-                "https://git-annex.branchable.com"
+    def run_in_thread(cls, decorated_fun):
+        @functools.wraps(decorated_fun)
+        def wrapper(*args, **kwargs):
+            thread = threading.Thread(
+                target=decorated_fun, args=args, kwargs=kwargs
             )
-        return s
+            logger.debug(
+                f"Starting {decorated_fun}(*{args!r}, **{kwargs!r}) in thread"
+            )
+            thread.start()
+            # logger.debug(
+            #     f"Waiting for {decorated_fun}(*{args!r}, **{kwargs!r}) thread to exit..."
+            # )
+            # thread.join()
 
-    def make_metadata_cache(self, path, cache=None):
+        return wrapper
+
+    @classmethod
+    def only_for_unique_git_annex_repo(cls, decorated_fun):
+        @functools.wraps(decorated_fun)
+        def wrapper(*args, **kwargs):
+            items = args[-1]
+            logger.debug(
+                f"only_for_unique_git_annex_repo: {args = }, {kwargs = }, {items = }"
+            )
+            folders = set(
+                (
+                    f.get_location().get_path()
+                    if f.is_directory()
+                    else os.path.dirname(f.get_location().get_path())
+                )
+                for f in items
+            )
+            folder_uuids = {d: cls.get_git_annex_uuid(d) for d in folders}
+            uuids = set(folder_uuids.values())
+            if not (len(uuids) == 1 and all(uuids)):
+                logger.info(
+                    f"Not exactly ONE unique git-annex repo selected: "
+                    f"{folder_uuids = }"
+                )
+                return []
+            info = dict(
+                cwd=(cwd := next(iter(folder_uuids))),
+                repo=cls.get_git_repo(cwd),
+                uuid=next(iter(uuids), None),
+            )
+            return decorated_fun(*args, info=info, **kwargs)
+
+        return wrapper
+
+    @classmethod
+    def make_metadata_cache(cls, path, cache=None):
         """
         Make/Update a ``dict[file][field] = {'val3','val2'}``
         """
         notification = Notify.Notification.new(
             _("Git Annex Thunar Plugin"), None, "git-annex"
         )
-        self.notification = notification
+        cls.notification = notification
         if cache is None:
-            cache = collections.defaultdict(
-                lambda: collections.defaultdict(set)
-            )
+            cache = defaultdict(lambda: defaultdict(set))
         notification.update(
             _("Git Annex Thunar Plugin"),
             "⏳ " + _("Determining amount of files..."),
@@ -82,7 +127,7 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
         )
         notification.show()
         for n_files, f in enumerate(
-            self.subprocess(
+            cls.subprocess(
                 subprocess.Popen,
                 ["git", "-C", path, "ls-files"],
                 stdout=subprocess.PIPE,
@@ -91,7 +136,7 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
         ):
             pass
         try:
-            git_annex = self.subprocess(
+            git_annex = cls.subprocess(
                 subprocess.Popen,
                 ["git", "-C", path, "annex", "metadata", "--json"],
                 stdin=subprocess.DEVNULL,
@@ -167,11 +212,12 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
             console.print(Pretty(dict(cache), max_length=20))
         return cache
 
-    def rebuild_metadata_cache(self, path):
-        uuid = self.get_git_annex_uuid(path)
+    @classmethod
+    def rebuild_metadata_cache(cls, path):
+        uuid = cls.get_git_annex_uuid(path)
         if not uuid:
             return METADATA_CACHE[uuid]
-        METADATA_CACHE[uuid] = self.make_metadata_cache(path=path)
+        METADATA_CACHE[uuid] = cls.make_metadata_cache(path=path)
 
     @classmethod
     def subprocess(cls, fun, cmd, *args, **kwargs):
@@ -412,27 +458,45 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
             ],
         ).rstrip("\r\n")
 
-    def get_file_menu_items(self, window, items):
-        folders = set(
-            (
-                f.get_location().get_path()
-                if f.is_directory()
-                else os.path.dirname(f.get_location().get_path())
-            )
-            for f in items
+
+class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider, GitAnnexHelper):
+    def __init__(self):
+        logger.debug(f"{type(self).__name__} is initialized")
+
+    @property
+    def counter(self):
+        try:
+            return self._counter
+        except AttributeError:
+            self._counter = collections.Counter()
+        return self._counter
+
+    @classmethod
+    def name(cls):
+        s = _("Git Annex Context-Menu")
+        if not shutil.which("git-annex"):
+            s = _("[Unavailable]") + " " + s
+        return s
+
+    @classmethod
+    def description(cls):
+        s = _(
+            "This plugin adds a context menu item "
+            "for managing Git Annex repositories."
         )
-        folder_uuids = {d: self.get_git_annex_uuid(d) for d in folders}
-        uuids = set(folder_uuids.values())
-        if not (len(uuids) == 1 and all(uuids)):
-            logger.info(
-                f"Not exactly ONE unique git-annex repo selected: "
-                f"{folder_uuids = }"
+        if not shutil.which("git-annex"):
+            s += " " + _(
+                "Install Git Annex to use this plugin: "
+                "https://git-annex.branchable.com"
             )
-            return []
-        uuid = next(iter(uuids), None)
-        cwd = next(iter(folder_uuids))
+        return s
+
+    @GitAnnexHelper.only_for_unique_git_annex_repo
+    def get_file_menu_items(self, window, items, info):
+        uuid = info["uuid"]
+        cwd = info["cwd"]
         logger.debug(f"Will operate in {cwd = }")
-        repo = self.get_git_repo(cwd)
+        repo = info["repo"]
         branch = self.get_git_branch(repo)
         branches = self.get_git_branches(folder=repo, all=False, sensible=True)
         in_view = re.search(r"^views/.+", branch, flags=re.IGNORECASE)
@@ -681,3 +745,213 @@ class GitAnnexSubmenu(GObject.GObject, Thunarx.MenuProvider):
 
     def get_folder_menu_items(self, window, folder):
         return self.get_file_menu_items(window, [folder])
+
+
+class GitAnnexProperties(
+    GObject.GObject, Thunarx.PropertyPageProvider, GitAnnexHelper
+):
+    def __init__(self):
+        logger.debug(f"{type(self).__name__} is initialized")
+
+    @property
+    def ui(self):
+        try:
+            return self._ui
+        except AttributeError:
+            self._ui = Gtk.Builder()
+            self._ui.set_translation_domain(l10n.GETTEXT_DOMAIN)
+            self._ui.add_from_file(
+                pkg_resources.resource_filename(
+                    "thunar_plugins.ui", "git-annex-properties.glade"
+                )
+            )
+
+            handlers = {
+                "readMetadata": self.readMetadata,
+                "editMetadata": self.editMetadata,
+                "metadataTextChanged": self.metadataTextChanged,
+                "stopReadMetadata": self.stopReadMetadata,
+            }
+            self._ui.connect_signals(handlers)
+        return self._ui
+
+    def stopReadMetadata(self, *args, **kwargs):
+        self._stopReadMetadata = True
+
+    def editMetadata(self, widget, *args, **kwargs):
+        field_entry = self.ui.get_object("git_annex_metadata_field_entry")
+        value_entry = self.ui.get_object("git_annex_metadata_value_entry")
+        if not (field := field_entry.props.text):
+            return
+        if not (value := value_entry.props.text):
+            return
+        optmakers = dict(
+            git_metadata_set_button=lambda f, v: ["--set", f"{f}={v}"],
+            git_metadata_add_button=lambda f, v: ["--set", f"{f}+={v}"],
+            git_metadata_setdefault_button=lambda f, v: ["--set", f"{f}?={v}"],
+            git_metadata_remove_button=lambda f, v: ["--set", f"{f}-={v}"],
+            git_metadata_clear_button=lambda f, v: ["--remove", f],
+        )
+        for btnid, optmaker in optmakers.items():
+            if widget is self.ui.get_object(btnid):
+                cmdparts = (
+                    [
+                        "git",
+                        "-C",
+                        self.info["cwd"],
+                        "annex",
+                        "metadata",
+                        "--force",
+                    ]
+                    + optmaker(field, value)
+                    + self.paths
+                )
+                break
+        if not cmdparts:
+            logger.error(
+                f"Don't know what to do with metadata {field = !r} {value = !r} from {widget = }"
+            )
+            return
+        self.ui.get_object(
+            "git_annex_metadata_edit_spinner"
+        ).props.active = True
+        with self.subprocess(subprocess.Popen, cmdparts) as proc:
+            while (returncode := proc.poll()) is None:
+                self.flush_gtk()
+        image = self.ui.get_object("git_annex_metadata_edit_image")
+        self.ui.get_object(
+            "git_annex_metadata_edit_spinner"
+        ).props.active = False
+        if returncode == 0:
+            image.props.icon_name = "dialog-ok"
+            image.props.tooltip_text = None
+        else:
+            image.props.icon_name = "dialog-error"
+            image.props.tooltip_text = _(
+                "Error setting Git Annex metadata. "
+                "The following command returned code {returncode}:\n\n{cmdline}"
+            ).format(returncode=returncode, cmdline=shlex.join(cmdline))
+
+    def metadataTextChanged(self, *args, **kwargs):
+        """Make button sensitive iff text is set"""
+        field_entry = self.ui.get_object("git_annex_metadata_field_entry")
+        value_entry = self.ui.get_object("git_annex_metadata_value_entry")
+        logger.debug(
+            f"metadataTextChanged(): {field_entry.props.text = !r}, {value_entry.props.text = !r}"
+        )
+        field_ok = re.search(
+            r"^[A-Z0-9_.-]+$", field_entry.props.text, flags=re.IGNORECASE
+        )
+        field_entry.props.primary_icon_name = (
+            "gtk-dialog-error"
+            if not field_ok and field_entry.props.text
+            else None
+        )
+        for button in map(
+            self.ui.get_object,
+            (
+                "git_metadata_set_button",
+                "git_metadata_add_button",
+                "git_metadata_setdefault_button",
+                "git_metadata_remove_button",
+                "git_metadata_clear_button",
+            ),
+        ):
+            button.props.sensitive = value_entry.props.text and field_ok
+        self.ui.get_object(
+            "git_annex_metadata_edit_image"
+        ).props.icon_name = None
+        self.ui.get_object(
+            "git_annex_metadata_edit_spinner"
+        ).props.active = False
+
+    @GitAnnexHelper.not_in_parallel
+    def readMetadata(self, widget, *args, **kwargs):
+        treestore = self.ui.get_object("git_annex_metadata_treestore")
+        liststore = self.ui.get_object("git_annex_metadata_paths_liststore")
+        treestore_filled = bool(treestore.iter_children(None))
+        if treestore_filled and widget is not self.ui.get_object(
+            "git_annex_read_metadata_button"
+        ):
+            # logger.debug(
+            #     f"Not updating metadata because treestore is "
+            #     "already filled and button was not clicked."
+            # )
+            return
+        self._stopReadMetadata = False
+        button = self.ui.get_object("git_annex_read_metadata_button")
+        button.props.sensitive = False
+        pbar = self.ui.get_object("git_annex_metadata_read_progressbar")
+        treeview = self.ui.get_object("git_annex_metadata_treeview")
+        treeview.props.sensitive = False
+        metadata = defaultdict(Counter)
+        n_files = 0
+        with self.subprocess(
+            subprocess.Popen,
+            ["git", "-C", self.info["cwd"], "annex", "metadata", "--json"]
+            + self.paths,
+            stdout=subprocess.PIPE,
+        ) as git_annex:
+            for n_files, metadata_json_line in enumerate(
+                git_annex.stdout, start=1
+            ):
+                if self._stopReadMetadata:
+                    logger.info(f"🛑 Stop reading metadata")
+                    git_annex.terminate()
+                    break
+                pbar.props.text = _("read metadata from {n} files").format(
+                    n=n_files
+                )
+                pbar.pulse()
+                self.flush_gtk()
+                try:
+                    metadata_json = json.loads(metadata_json_line)
+                except json.JSONDecodeError as e:
+                    logger.error(
+                        f"git-annex returned weird JSON {metadata_json_line!r}: {e}"
+                    )
+                    continue
+                if path := metadata_json.get("file"):
+                    liststore.append([path])
+                for field, values in metadata_json.get("fields", {}).items():
+                    if not field.endswith("lastchanged"):
+                        for value in values:
+                            metadata[field][value] += 1
+
+        pbar.props.fraction = 1
+        self.flush_gtk()
+        with console.status("Filling treeview..."):
+            treestore.clear()
+            for field, values in sorted(metadata.items()):
+                fielditer = treestore.append(None, (field, None))
+                for value, count in sorted(values.items()):
+                    valueiter = treestore.append(
+                        fielditer,
+                        (_("{n} files").format(n=count) + " ⮕ ", value),
+                    )
+            treeview.expand_all()
+        if logger.getEffectiveLevel() <= logging.DEBUG:
+            logger.debug(f"metadata from {n_files} selected files: ")
+            console.log(dict(metadata))
+        button.props.sensitive = True
+        treeview.props.sensitive = True
+
+    @GitAnnexHelper.only_for_unique_git_annex_repo
+    def get_property_pages(self, files, info):
+        if hasattr(self, "_ui"):
+            # necessary for some reason as the properties page is empty
+            # otherwise 🤷
+            del self._ui
+        self.paths = [str(f.get_location().get_path()) for f in files]
+        self.info = info
+        if logger.getEffectiveLevel() <= logging.DEBUG:
+            logger.debug(f"paths for git annex properties page:")
+            console.log(self.paths)
+
+        page_toplevel_box = self.ui.get_object(
+            "git_annex_metadata_properties_page"
+        )
+
+        git_annex_page = Thunarx.PropertyPage.new(_("Git Annex"))
+        git_annex_page.add(page_toplevel_box)
+        return [git_annex_page]
